import pandas as pd

def get_run_details(comparison_obj):
    """
    Generates a summary DataFrame about the comparison run.
    """
    try:
        data = {
            "run_id": [comparison_obj.run_id],
            "Stream": [comparison_obj.stream],
            "new_report_workspace": [f"{comparison_obj.report_new.report_name}_workspace_{comparison_obj.report_new.workspace_name}"],
            "old_report_workspace": [f"{comparison_obj.report_old.report_name}_workspace_{comparison_obj.report_old.workspace_name}"],
            "new_report_refresh_date": [str(comparison_obj.report_new.last_modified_date)],
            "old_report_refresh_date": [str(comparison_obj.report_old.last_modified_date)]
        }
        return pd.DataFrame(data)
    except Exception as e:
        print(f"Error generating run details: {e}")
        return pd.DataFrame()

def get_raw_table_details(comparison_obj):
    """
    Returns all table comparison details with a run ID.
    """
    try:
        df = comparison_obj._all_tables.copy()
        df["run_id"] = comparison_obj.run_id
        return df
    except Exception as e:
        print(f"Error retrieving raw table details: {e}")
        return pd.DataFrame()

def get_raw_measure_details(comparison_obj):
    """
    Returns all measure comparison details with a run ID.
    """
    try:
        df = comparison_obj._all_measures.copy()
        df["run_id"] = comparison_obj.run_id
        return df
    except Exception as e:
        print(f"Error retrieving raw measure details: {e}")
        return pd.DataFrame()


import time
import pandas as pd
from .comparers import TableComparer, MeasureComparer, ColumnComparer
from .validators import TableValidator, MeasureValidator, ColumnValidator
from .utils import get_run_details, get_raw_table_details, get_raw_measure_details
from .report import FabricAnalyticsReport 
from sempy import fabric as sfabric
from .token_provider import UserCredAuth

import logging 
import threading
logger = logging.getLogger(__name__)

class ReportComparison:
    """
    Orchestrates the comparison and validation of two FabricAnalyticsReport objects.
    """

    def __init__(
            self, 
            report_new : FabricAnalyticsReport, 
            report_old : FabricAnalyticsReport, 
            stream: str, 
            threshold_score: float = 80.0
            ):
        """
        Initializes the ReportComparison object.

        Args:
            report_new (FabricAnalyticsReport): The new version of the report.
            report_old (FabricAnalyticsReport): The old version of the report.
            stream (str): A label for the comparison stream.
            threshold_score (float, optional): The fuzzy matching threshold. Defaults to 90.0.
        """
        self.report_new = report_new
        self.report_old = report_old
        self.stream = str.lower(str.strip(stream))
        self.threshold_score = threshold_score
        self.run_id = str(int(time.time()))

        # Comparers
        self._table_comparer = TableComparer(self.report_new.tables, self.report_old.tables, self.threshold_score)
        self._all_tables = self._table_comparer.compare()
        self.common_tables = self._all_tables[self._all_tables['origin'] == 'both']
        
        self._column_comparer = ColumnComparer(self.report_new.columns, self.report_old.columns, self.common_tables, self.threshold_score)
        self._all_columns = self._column_comparer.compare()
        self.common_columns = self._all_columns[self._all_columns['origin'] == 'both']
        
        self._measure_comparer = MeasureComparer(self.report_new.measures, self.report_old.measures, self.threshold_score)
        self._all_measures = self._measure_comparer.compare()
        self.common_measures = self._all_measures[self._all_measures['origin'] == 'both']
        
        # Results initialized to None
        self.TableValidationResults = None
        self.MeasureValidationResults = None
        self.ColumnValidationResults = None
        self.RunDetails = get_run_details(self)
        self.RawTables = None
        self.RawMeasures = None

    def run_table_validation(self, margin_of_error: float = 5.0):
        """
        Runs table row count validation.
        
        Args:
            margin_of_error (float, optional): Allowed percentage difference. Defaults to 5.0.
        """
        try:
            logger.info("Starting table validation...")
            start_time = time.time()
            self.table_validator = TableValidator(self.report_new, self.report_old, self.common_tables, self.run_id, self.stream)
            self.TableValidationResults = self.table_validator.validate_row_counts(margin_of_error)
            self.RawTables = get_raw_table_details(self)
            end_time = time.time()
            logger.info(f"Table validation completed in {end_time - start_time:.2f} seconds")
        except Exception as e:
            logger.error(f"Error in table row count validation: {str(e)}")
            self.TableValidationResults = self.RawTables = pd.DataFrame()
    
    def run_column_validation(self, margin_of_error: float = 5.0):
        """
        Runs column level validations.
        
        Args:
            margin_of_error (float, optional): Allowed percentage difference. Defaults to 5.0.
        """
        try:
            logger.info("Starting column validation...")
            start_time = time.time()
            self.column_validator = ColumnValidator(self.report_new, self.report_old, self.common_columns, self.run_id, self.stream)
            self.ColumnValidationResults = self.column_validator.validate_distinct_counts(margin_of_error)
            end_time = time.time()
            logger.info(f"Column validation completed in {end_time - start_time:.2f} seconds")
        except Exception as e:
            logger.error(f"Error in column validation: {str(e)}")
            self.ColumnValidationResults = pd.DataFrame()

    def run_measure_validation(self, margin_of_error: float = 5.0):
        """
        Runs measure level validation.
        
        Args:
            margin_of_error (float, optional): Allowed percentage difference. Defaults to 5.0.
        """
        try:
            logger.info("Starting measure validation...")
            start_time = time.time()
            self.measure_validator = MeasureValidator(self.report_new, self.report_old, self.common_measures, self.run_id, self.stream)
            self.MeasureValidationResults = self.measure_validator.validate_measure_values(margin_of_error)
            self.RawMeasures = get_raw_measure_details(self)
            end_time = time.time()
            logger.info(f"Measure validation completed in {end_time - start_time:.2f} seconds")
        except Exception as e:
            logger.error(f"Error in measure validation: {str(e)}")
            self.MeasureValidationResults = self.RawMeasures = pd.DataFrame()

    def run_all_validations(self, margin_of_error: float = 5.0) -> dict:
        """
        Runs all validation functions (table, column, measure).
        
        Args:
            margin_of_error (float, optional): Allowed percentage difference. Defaults to 5.0.
        """
        logger.info("Starting all validations ...")
        start_time_parallel = time.time()
        threads = []

        # Creating a thread for each validation function
        table_thread = threading.Thread(target=self.run_table_validation, args=(margin_of_error,))
        column_thread = threading.Thread(target=self.run_column_validation, args=(margin_of_error,))
        measure_thread = threading.Thread(target=self.run_measure_validation, args=(margin_of_error,))

        # Adding threads to a list
        threads.append(table_thread)
        threads.append(column_thread)
        threads.append(measure_thread)

        # Starting all threads
        for thread in threads:
            thread.start()

        # Waiting for all threads to complete
        for thread in threads:
            thread.join()
        end_time_parallel = time.time()
        results = {
            "Measure Validation results" : self.MeasureValidationResults,
            "Table Validation results": self.TableValidationResults,
            "Column Validation results": self.ColumnValidationResults
        }
        logger.info(f"All validations completed in {end_time_parallel - start_time_parallel:.2f} seconds")
        return results

# def ReportCompare(ReportsToCompare: list, Stream: str, ExplicitToken:str|None = None) -> ReportComparison:
#     """
#     Compare two reports across workspaces.

#     Args:
#         ReportsToCompare (dict): A dictionary with the following keys:
#             - 'old_report' (str): Name of the old report.
#             - 'old_report_workspace' (str): Workspace Name of the old report.
#             - 'new_report' (str): Name of the new report.
#             - 'new_report_workspace' (str): Workspace Name of the new report.
#         Stream (str): A label or stream name used for tagging the comparison run.
#         ExplicitToken (optional): A token to use for authentication. If None, a default token provider is used.
#     Returns:
#         str: A summary or status string of the comparison process.
#     """    
#     if ExplicitToken:
#         sfabric._token_provider._get_token = UserCredAuth(ExplicitToken) # type: ignore

#     OldReport = FabricAnalyticsReport(ReportsToCompare[0]['old_report'],ReportsToCompare[0]['old_report_workspace'])
#     NewReport = FabricAnalyticsReport(ReportsToCompare[1]['new_report'],ReportsToCompare[1]['new_report_workspace'])

#     Compare = ReportComparison(NewReport, OldReport, Stream)
#     return Compare

def ReportCompare(
        OldReport: str, 
        OldReportWorkspace: str, 
        NewReport: str, 
        NewReportWorkspace: str,
        Stream: str, 
        ExplicitToken:str|None = None
    ) -> ReportComparison:
    """
    Compare two reports across workspaces.

    Args:
       'old_report' (str): Name of the old report.
       'old_report_workspace' (str): Workspace Name of the old report.
       'new_report' (str): Name of the new report.
       'new_report_workspace' (str): Workspace Name of the new report.
        Stream (str): A label or stream name used for tagging the comparison run.
        ExplicitToken (optional): A token to use for authentication. If None, a default token provider is used.
    Returns:
        ReportComparison Object: This can be futher used to run data validations.
    """    
    if ExplicitToken:
        sfabric._token_provider._get_token = UserCredAuth(ExplicitToken) # type: ignore

    old_report = FabricAnalyticsReport(OldReport,OldReportWorkspace)
    new_report = FabricAnalyticsReport(NewReport,NewReportWorkspace)

    Compare = ReportComparison(old_report, new_report, Stream)
    return Compare




import pandas as pd
from thefuzz import fuzz, process
import logging 
logger = logging.getLogger(__name__)

class BaseComparer:
    def __init__(self, new_df : pd.DataFrame, old_df : pd.DataFrame, threshold_score : float):
        self.new_df = new_df.copy()
        self.old_df = old_df.copy()
        self.threshold_score = threshold_score

    def _fuzzy_match(self, new_items: pd.DataFrame, old_items: pd.DataFrame, match_col: str) -> pd.DataFrame:
        """
        Performs a more efficient fuzzy match of new_items to old_items on a given column.
        
        Args:
            new_items (pd.DataFrame): DataFrame of new items to be matched.
            old_items (pd.DataFrame): DataFrame of old items to be matched against.
            match_col (str): The column name to perform the fuzzy match on.

        Returns:
            pd.DataFrame: A DataFrame with matched new and old items, plus any unmatched items.
        """
        matched_results = []
        
        # Create a temporary index for old_items to easily reference them
        old_items_indexed = old_items.reset_index(drop=True).copy()
        old_choices_list = old_items_indexed[match_col].astype(str).tolist()
        
        # Keep track of which old_items have been matched
        # Using a set of indices for efficient lookup
        matched_old_indices = set() 
        
        # Store all potential matches above threshold, then sort
        all_potential_matches = [] # (new_item_row, old_item_index, score)

        for new_idx, new_row in new_items.iterrows():
            query_string = str(new_row[match_col])
            
            # Use process.extract to get ALL matches above the score_cutoff
            # This returns a list of (match_string, score, index_in_choices_list)
            results = process.extractOne(
                query=query_string,
                choices=old_choices_list,
                scorer=fuzz.ratio, 
                score_cutoff=round(self.threshold_score)
            )
            if results is None:
                continue

            if len(results) == 3:
                match_str, score, old_list_idx  =results
            else:
                match_str, score = results
                old_list_idx = old_choices_list.index(match_str)

            all_potential_matches.append({
                    'new_idx': new_idx,
                    'old_list_idx': old_list_idx,
                    'score': score
                })

        # Sort potential matches by score in descending order
        # This ensures we prioritize the best possible match globally
        all_potential_matches.sort(key=lambda x: x['score'], reverse=True)

        matched_new_indices = set()

        for match_info in all_potential_matches:
            new_idx = match_info['new_idx']
            old_list_idx = match_info['old_list_idx']
            score = match_info['score']

            # If both the new item and the old item haven't been matched yet
            if new_idx not in matched_new_indices and old_list_idx not in matched_old_indices:
                new_row = new_items.loc[new_idx]
                old_row = old_items_indexed.loc[old_list_idx]
                
                combined = {f"{col}_new": new_row[col] for col in new_items.columns}
                combined.update({f"{col}_old": old_row[col] for col in old_items.columns})
                combined["best_score"] = score
                combined["origin"] = "both"
                matched_results.append(combined)
                
                matched_new_indices.add(new_idx)
                matched_old_indices.add(old_list_idx)
        
        # Add unmatched new items
        for new_idx, new_row in new_items.iterrows():
            if new_idx not in matched_new_indices:
                combined = {f"{col}_new": new_row[col] for col in new_items.columns}
                combined.update({f"{col}_old": None for col in old_items.columns}) # Use old_items.columns for consistent column names
                combined["best_score"] = 0
                combined["origin"] = "new"
                matched_results.append(combined)
                
        # Add unmatched old items
        for old_list_idx, old_row in old_items_indexed.iterrows():
            if old_list_idx not in matched_old_indices:
                combined = {f"{col}_new": None for col in new_items.columns} # Use new_items.columns for consistent column names
                combined.update({f"{col}_old": old_row[col] for col in old_items.columns})
                combined["best_score"] = 0 # type: ignore
                combined["origin"] = "old" # type: ignore
                matched_results.append(combined)

        result_df = pd.DataFrame(matched_results)
        if result_df.empty:
            logger.warning("No matching entries were found.")
        return result_df


class TableComparer(BaseComparer):
    def compare(self) -> pd.DataFrame:
        try:
            return self._fuzzy_match(self.new_df, self.old_df, 'table_name')
        except Exception as e:
            logger.error(f"Error in table name comparison: {str(e)}")
            self.TableValidationResults = self.RunDetails = self.RawTables = pd.DataFrame()
            return pd.DataFrame()
        
class MeasureComparer(BaseComparer):
    def compare(self) -> pd.DataFrame:
        try:
            return self._fuzzy_match(self.new_df, self.old_df, 'field_name')
        except Exception as e:
            logger.error(f"Error in measure name comparison: {str(e)}")
            return pd.DataFrame()

class ColumnComparer(BaseComparer):
    def __init__(self, 
                 new_df : pd.DataFrame, 
                 old_df : pd.DataFrame, 
                 matched_tables_both : pd.DataFrame, 
                 threshold_score : float
        ):
        super().__init__(new_df, old_df, threshold_score)
        self.matched_tables_both = matched_tables_both

    def compare(self) -> pd.DataFrame:
        all_matched_columns = []
        
        matched_tables_both = self.matched_tables_both.copy()
        
        for _, table_row in matched_tables_both.iterrows():
            new_table_name = table_row['table_name_new']
            old_table_name = table_row['table_name_old']
            
            new_cols = self.new_df[self.new_df['table_name'] == new_table_name]
            old_cols = self.old_df[self.old_df['table_name'] == old_table_name]
            
            matched_columns = self._fuzzy_match(new_cols, old_cols, 'field_name')
            matched_columns[matched_columns['origin'] == 'both'] 
            all_matched_columns.append(matched_columns)
            
        if not all_matched_columns:
            return pd.DataFrame()

        return pd.concat(all_matched_columns, ignore_index=True)


# knnpy/__init__.py
"""
The knnpy package for Power BI report comparison and validation.
"""

# Import key classes/functions to make them directly accessible from the package
from .base import (
    ReportCompare,
    ReportComparison
)
from .token_provider import initializeToken
from .report import FabricAnalyticsReport
from .utils import (
    get_raw_measure_details,
    get_raw_table_details,
    get_run_details
)

# Define __all__ for explicit imports
__all__ = [
    "FabricAnalyticsReport",
    "initializeToken",
    "ReportCompare"
]

import logging
logger = logging.getLogger(__name__)


# logger.setLevel(logging.INFO)

# console_handler = logging.StreamHandler()
# console_handler.setLevel(logging.INFO)
# formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
# console_handler.setFormatter(formatter)
# if not logger.handlers:
#     logger.addHandler(console_handler)


import pandas as pd
import sempy.fabric as sfabric
import logging 
logger = logging.getLogger(__name__)

class FabricAnalyticsReport:
    """
    Represents a Power BI report in Microsoft Fabric, providing methods
    to extract metadata like tables, columns, and measures.
    """
    def __init__(self, report_name: str, workspace_name: str) -> None:
        """
        Initializes the FabricAnalyticsReport object.

        Args:
            report_name (str): The name of the Power BI report.
            workspace_name (str): The name of the workspace containing the report.
        
        Raises:
            RuntimeError: If workspace or dataset resolution fails.
        """
        self.report_name = report_name
        self.workspace_name = workspace_name

        try:
            self.workspaceid = sfabric.resolve_workspace_id(workspace_name)
        except Exception as e:
            logger.error(f"Failed to resolve workspace ID for '{workspace_name}': {e}")
            raise RuntimeError(f"Workspace resolution failed for: {workspace_name}")

        try:
            self.datasetid = sfabric.resolve_dataset_id(workspace=self.workspaceid, dataset_name=report_name)
        except Exception as e:
            logger.error(f"Failed to resolve dataset ID for report '{report_name}' in workspace '{workspace_name}': {e}")
            raise RuntimeError(f"Dataset resolution failed for: {report_name} in {workspace_name}")

        self.measures = self.__getMeasures()
        self.columns = self.__getAttributes()
        self.tables = self.__getTables()
        self.last_modified_date = self.__getLastModifiedDate()

    def __getMeasures(self) -> pd.DataFrame:
        """
        Retrieves all measures from the dataset.
        
        Returns:
            pd.DataFrame: A DataFrame of measures with standardized column names.
        """
        try:
            df = sfabric.list_measures(dataset=self.datasetid, workspace=self.workspaceid)

            if df is None or df.empty:
                logger.warning("Measure list is empty or dataset might not contain any measures.")
                return pd.DataFrame()
            return df[['Table Name', 'Measure Name', 'Measure Expression', 'Measure Data Type']].rename(columns={
                'Table Name': 'table_name',
                'Measure Name': 'field_name',
                'Measure Expression': 'expression',
                'Measure Data Type': 'data_type'
            }) 
        except Exception as e:
            logger.error(f"Error fetching measures for report '{self.report_name}' in workspace '{self.workspace_name}': {e}")
            return pd.DataFrame()

    def __getAttributes(self) -> pd.DataFrame:
        """
        Retrieves all columns (attributes) from the dataset.

        Returns:
            pd.DataFrame: A DataFrame of columns with standardized column names.
        """
        try:
            df = sfabric.list_columns(dataset=self.datasetid, workspace=self.workspaceid)

            if df is None or df.empty:
                logger.warning("Attribute list is empty or dataset might not contain any columns.")
                return pd.DataFrame()
            return df[['Table Name', 'Column Name', 'Data Type']].rename(columns={
                'Table Name': 'table_name',
                'Column Name': 'field_name',
                'Data Type': 'data_type'
            })
        except Exception as e:
            logger.error(f"Error fetching attributes (columns) for report '{self.report_name}' in workspace '{self.workspace_name}': {e}")
            return pd.DataFrame()

    def __getTables(self) -> pd.DataFrame:
        """
        Retrieves all tables from the dataset.

        Returns:
            pd.DataFrame: A DataFrame of tables.
        """
        try:
            df = sfabric.list_tables(dataset=self.datasetid, workspace=self.workspaceid)

            if df is None or df.empty:
                logger.warning("Table list is empty or dataset might not contain any tables.")
                return pd.DataFrame()
            return df[['Name']].rename(columns={'Name': 'table_name'})
        except Exception as e:
            logger.error(f"Error fetching tables for report '{self.report_name}' in workspace '{self.workspace_name}': {e}")
            return pd.DataFrame()

    def __getLastModifiedDate(self) -> str:
        """
        Retrieves the last refresh date of the dataset.

        Returns:
            str: The last modified date as a formatted string.
        """
        try:
            refresh_details = sfabric.list_refresh_requests(dataset=self.datasetid, workspace=self.workspaceid)
            if refresh_details.empty:
                logger.warning(f"No refresh history found. Using placeholder date for report '{self.report_name}' in workspace '{self.workspace_name}'.")
                return '01/01/1999 01:01'
            refresh_end = refresh_details[refresh_details['Status'] == 'Completed'] \
                .sort_values('End Time', ascending=False).iloc[0]['End Time']
            return refresh_end.strftime('%m/%d/%Y %H:%M') if refresh_end else '01/01/1999 01:01'
        except Exception as e:
            logger.error(f"Error retrieving last modified date: {e}, putting a placeholder - 01/01/1999")
            return '01/01/1999 01:01'


from typing import  Literal
import jwt.utils
import sempy.fabric as sfabric
import jwt

class UserCredAuth:
    def __init__(self,token: str)->None:
        self.acces_token = token

    def __call__(self,audience: Literal["pbi", "storage"] = "pbi") -> str:
            return self.acces_token


def initializeToken(ExplicitToken):
    sfabric._token_provider._get_token = UserCredAuth(ExplicitToken) # type: ignore



import pandas as pd
import numpy as np
import sempy.fabric as sfabric
from .report import FabricAnalyticsReport
import logging 
logger = logging.getLogger(__name__)

class BaseValidator:
    def __init__(
            self, 
            report_new: FabricAnalyticsReport, 
            report_old: FabricAnalyticsReport, 
            all_items: pd.DataFrame, 
            run_id: str, 
            stream: str
        ):
        self.report_new = report_new
        self.report_old = report_old
        self.all_items = all_items
        self.run_id = run_id
        self.stream = stream
    
    def _validate_value(self, new_val, old_val, margin_of_error) -> bool:

        if pd.isna(new_val) or pd.isna(old_val):
            return False

        lower_bound = old_val * (1 - margin_of_error / 100)
        upper_bound = old_val * (1 + margin_of_error / 100)

        return True if lower_bound <= new_val <= upper_bound else False
    
class MeasureValidator(BaseValidator):

    def validate_measure_values(self, margin_of_error=5.0):

        matched_measures = self.all_items[
            (self.all_items['origin'] == 'both') &
            (self.all_items['data_type_new'].str.strip().str.lower() == 'int64') &
            (self.all_items['data_type_old'].str.strip().str.lower() == 'int64')
        ][[
            'field_name_new',
            'field_name_old',
            'data_type_new',
            'data_type_old',
            'expression_old',
            'expression_new',
            'origin',
            'best_score'
        ]].dropna(subset=['field_name_new', 'field_name_old'])

        if matched_measures.empty:
            return pd.DataFrame()

        try:
            new_eval = sfabric.evaluate_measure(
                dataset=self.report_new.datasetid,
                workspace=self.report_new.workspaceid,
                measure=list(matched_measures['field_name_new'])
            )
            old_eval = sfabric.evaluate_measure(
                dataset=self.report_old.datasetid,
                workspace=self.report_old.workspaceid,
                measure=list(matched_measures['field_name_old'])
            )
            new_eval_pivot = new_eval.melt(var_name='field_name_new', value_name='new_report_value')
            old_eval_pivot = old_eval.melt(var_name='field_name_old', value_name='old_report_value')

            merged = pd.merge(matched_measures, new_eval_pivot, on='field_name_new', how='left')
            merged = pd.merge(merged, old_eval_pivot, on='field_name_old', how='left')
            
            merged['value_difference'] = merged['new_report_value'] - merged['old_report_value']
            merged['value_difference_percent'] = np.where(
                (merged['old_report_value'] == 0) |
                (merged['old_report_value'].isna()) |
                (merged['new_report_value'].isna()),
                '<NA>',
                (((merged['new_report_value'] - merged['old_report_value']) * 100) / merged['old_report_value'])
                .round(2).astype(str) + '%'
            )
            merged.loc[(merged['value_difference'] == 0) & (merged['value_difference_percent'] == '<NA>'),'value_difference_percent'] = '0.0%'
            merged['is_value_similar'] = merged.apply(
                lambda row: self._validate_value(row['new_report_value'], row['old_report_value'], margin_of_error), axis=1
            )
            merged['is_data_type_same'] = merged['data_type_new'] == merged['data_type_old']
            merged['is_expression_same'] = merged['expression_new'] == merged['expression_old']
            merged['run_id'] = self.run_id
            merged['Stream'] = self.stream

            final_cols = [
                'run_id', 'Stream', 'field_name_new', 'field_name_old' , 'best_score', 'origin',
                'new_report_value', 'old_report_value',
                'value_difference', 'value_difference_percent',
                'is_value_similar', 'is_data_type_same', 'is_expression_same'
            ]
            return merged[final_cols].astype(str).fillna('Unknown')
        except Exception as e:
            print(f"Error in measure validation: {e}")
            return pd.DataFrame()

class TableValidator(BaseValidator):

    def _get_table_row_count(self, report, table_name) -> int:
        query = f'EVALUATE ROW("RowCount", COUNTROWS(\'{table_name}\'))'
        try:
            df = sfabric.evaluate_dax(dataset=report.datasetid, workspace=report.workspaceid, dax_string=query)
            colname = df.columns[0]
            value = df[colname].iloc[0]
            return 0 if pd.isna(value) else int(value)
        except Exception as e:
            logger.error(f"Error fetching row count for table : {e}")
            return -1

    def validate_row_counts(self, margin_of_error=5.0) -> pd.DataFrame:
        results = []
        matched_tables = self.all_items

        for _, row in matched_tables.iterrows():
            table_new = row['table_name_new']
            table_old = row['table_name_old']

            count_new = self._get_table_row_count(self.report_new, table_new)
            count_old = self._get_table_row_count(self.report_old, table_old)
            
            diff = count_new - count_old
            diff_pct = f"{((diff) / count_old * 100):.2f}%" if count_old != 0 else "∞%"

            results.append({
                'run_id': self.run_id,
                'Stream': self.stream,
                'table_name_new': table_new,
                'table_name_old': table_old,
                'best_score': row['best_score'],
                'origin': 'both',
                'row_count_new': count_new,
                'row_count_old': count_old,
                'row_count_difference': diff,
                'row_count_diff_percentage': diff_pct,
                'is_value_similar': self._validate_value(count_new, count_old, margin_of_error),
            })
        
        return pd.DataFrame(results)

class ColumnValidator(BaseValidator):

    def _read_table(self,table_name) -> None:
        self._new_tables_data = {}
        try:
            self._new_tables_data[table_name] = sfabric.read_table(
                dataset=self.report_new.datasetid,
                workspace=self.report_new.workspaceid,
                table=table_name
            )
        except Exception as e:
            logger.warning(f"Error reading new table '{table_name}': {e}")
            self._new_tables_data[table_name] = pd.DataFrame()

    def _generate_distinct_count_dax(self,columns: list[str],table_name) -> str:
        rows = []
        for col in columns:
            row = f'ROW("ColumnName", "{col}", "DistinctCount", COUNTROWS(SUMMARIZE(\'{table_name}\', \'{table_name}\'[{col}])))'
            rows.append(row)
        return  "EVALUATE\nUNION(\n    " + ",\n    ".join(rows) + "\n)"
        
    def _get_column_distinct_counts(self, table_name: str, columns: list[str], report: FabricAnalyticsReport) -> pd.DataFrame | None:
        """
        Executes a DAX query to fetch distinct counts for the given columns and table.
        """
        if not columns:
            logging.warning(f"No columns provided for table '{table_name}' in dataset {report.datasetid}.")
            return None
            
        dax_query = self._generate_distinct_count_dax(columns=columns, table_name=table_name)
        if dax_query is None:
            return None
            
        try:
            df = sfabric.evaluate_dax(
                dataset=report.datasetid,
                workspace=report.workspaceid,
                dax_string=dax_query
            )
            # Standardize column names from DAX result
            df.columns = ["ColumnName", "DistinctCount"]
            if(len(df) == 0):
                logging.warning(f"No row count returned for columns in table '{table_name}'")
            return df
        except Exception as e:
            logging.error(f"Error fetching distinct row count for columns in table '{table_name}': {e}")
            return None

    def validate_distinct_counts(self, margin_of_error: float = 5.0) -> pd.DataFrame:
        """
        Validates the distinct count of values for all common columns across all common tables.

        Args:
            margin_of_error (float): The allowed percentage difference between old and new counts.

        Returns:
            pd.DataFrame: A dataframe containing the validation results for each column,
                          including new/old counts and whether the validation passed.
        """
        validation_results = []
        
        # Find common tables based on the mapping in all_items
        common_tables = self.all_items[['table_name_new', 'table_name_old']].drop_duplicates()

        for _, row in common_tables.iterrows():
            table_name_new = row['table_name_new']
            table_name_old = row['table_name_old']

            logging.info(f"Validating distinct counts for table: '{table_name_new}' (new) vs '{table_name_old}' (old)")

            # Filter columns for the current table pair
            table_columns_map = self.all_items[
                (self.all_items['table_name_new'] == table_name_new) &
                (self.all_items['table_name_old'] == table_name_old)
            ]

            cols_new = table_columns_map['field_name_new'].unique().tolist()
            cols_old = table_columns_map['field_name_old'].unique().tolist()

            # Fetch distinct counts for both new and old reports
            df_new = self._get_column_distinct_counts(table_name_new, cols_new, self.report_new)
            df_old = self._get_column_distinct_counts(table_name_old, cols_old, self.report_old)

            if df_new is None or df_old is None:
                logging.warning(f"Could not retrieve data for one or both tables: {table_name_new}, {table_name_old}. Skipping.")
                continue

            # Rename columns for merging
            df_new = df_new.rename(columns={"ColumnName": "field_name_new", "DistinctCount": "distinct_count_new"})
            df_old = df_old.rename(columns={"ColumnName": "field_name_old", "DistinctCount": "distinct_count_old"})
            
            # Merge results based on the column mapping
            merged_df = pd.merge(table_columns_map, df_new, on="field_name_new")
            merged_df = pd.merge(merged_df, df_old, on="field_name_old")

            # Perform validation for each column
            merged_df['is_value_similar'] = merged_df.apply(
                lambda r: self._validate_value(r['distinct_count_new'], r['distinct_count_old'], margin_of_error),
                axis=1
            )
            merged_df['value_difference'] = merged_df['distinct_count_new'] - merged_df['distinct_count_old']
            merged_df['value_difference_percent'] = np.where(
                (merged_df['distinct_count_new'] == 0) |
                (merged_df['distinct_count_new'].isna()) |
                (merged_df['distinct_count_old'].isna()),
                '<NA>',
                (((merged_df['distinct_count_old'] - merged_df['distinct_count_new']) * 100) / merged_df['distinct_count_new'])
                .round(2).astype(str) + '%'
            )
            merged_df.loc[(merged_df['value_difference'] == 0) & (merged_df['value_difference_percent'] == '<NA>'),'value_difference_percent'] = '0.0%'
            merged_df['is_data_type_same'] = merged_df['data_type_new'] == merged_df['data_type_old']
            validation_results.append(merged_df)

        if not validation_results:
            return pd.DataFrame() 

        return pd.concat(validation_results, ignore_index=True)
    



from setuptools import setup, find_packages

setup(
    name='fabric_maverick',
    version='0.1.0.dev5',
    description='A Fabric Package for Semantic/Dataset validation',
    author='Nisarg Patel',
    author_email='nisargp@maqsoftware.com',
    packages=find_packages(),
    classifiers=[
        "Development Status :: 3 - Alpha",
        
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",

        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3 :: Only",
        "Framework :: Jupyter"
    ],
    python_requires=">=3.10,<3.12"
)

# python3 setup.py sdist
# python3 -m build 
# python3 -m  twine upload dist/* 

#“major.minor” versioning with developmental releases, release candidates and post-releases for minor corrections:
# 0.9
# 1.0.dev1
# 1.0.dev2
# 1.0.dev3
# 1.0.dev4
# 1.0c1
# 1.0c2
# 1.0
# 1.0.post1
# 1.1.dev1
# license-files = ["LICENSE", "AUTHORS.md"]
